# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np
import pytest
from monty.serialization import loadfn
from pydefect.analyzer.defect_structure_comparator import \
    DefectStructureComparator, SiteDiff
from pymatgen.core import Structure, IStructure, Lattice
from vise.tests.helpers.assertion import assert_msonable


@pytest.fixture
def structure_comparator():
    cu2o_perfect = IStructure(Lattice.cubic(5),
                              species=["Cu"] * 4 + ["O"] * 2,
                              coords=[[0.25, 0.25, 0.25],  # removed
                                      [0.25, 0.74, 0.74],  # removed
                                      [0.75, 0.75, 0.25],
                                      [0.75, 0.25, 0.75],
                                      [0, 0, 0],
                                      [0.5, 0.5, 0.5]])
    cu2o_defect = IStructure(Lattice.cubic(5),
                             species=["Cu"] * 3 + ["O"] * 2 + ["H"],
                             coords=[[0.25, 0.5, 0.5],   # inserted
                                     [0.76, 0.73, 0.24],
                                     [0.75, 0.25, 0.73],
                                     [0.05, 0.95, 0],
                                     [0.5, 0.5, 0.5],
                                     [0.25]*3])   # inserted

    return DefectStructureComparator(defect_structure=cu2o_defect,
                                     perfect_structure=cu2o_perfect)


@pytest.fixture
def structure_comparator_periodic_issue():
    cu2o_perfect = IStructure(Lattice.cubic(5),
                              species=["Cu"] * 4 + ["O"] * 2,
                              coords=[[0.25, 0.25, 0.25],
                                      [0.25, 0.75, 0.75],
                                      [0.75, 0.75, 0.25],
                                      [0.75, 0.25, 0.75],
                                      [0, 0, 0],
                                      [0.5, 0.5, 0.5]])
    # defect center is ([1.0, 1.0, 1.0] + [0.99, 0.99, 0.99]) / 2 = [0.995]*3
    cu2o_defect = IStructure(Lattice.cubic(5),
                             species=["Cu"] * 4 + ["O"] + ["H"],
                             coords=[[0.25, 0.25, 0.25],
                                     [0.25, 0.75, 0.75],
                                     [0.75, 0.75, 0.25],
                                     [0.75, 0.25, 0.75],
                                     [0.5, 0.5, 0.5],
                                     [0.99, 0.99, 0.99]])

    return DefectStructureComparator(defect_structure=cu2o_defect,
                                     perfect_structure=cu2o_perfect)


def test_atom_mapping_to_perfect(structure_comparator):
    assert structure_comparator.atom_mapping == {1: 2, 2: 3, 3: 4, 4: 5}
    assert structure_comparator.removed_indices == [0, 1]
    assert structure_comparator.inserted_indices == [0, 5]


def test_defect_structure_analyzer_defect_center(structure_comparator):
    actual = structure_comparator.defect_center_coord
    assert (actual == np.array([0.25, 0.435, 0.435])).all()


def test_defect_center_periodicity(structure_comparator_periodic_issue):
    actual = structure_comparator_periodic_issue.defect_center_coord
    assert (actual == np.array([0.995]*3)).all()


def test_neighboring_atom_indices(cubic_supercell):
    structure: Structure = cubic_supercell.copy()
    structure.pop(32)  # [0.25, 0, 0]
    structure.pop(0)  # [0, 0, 0]
    structure.append(species="Li", coords=[0.124, 0, 0])
    structure.to(filename="POSCAR")
    structure_analyzer = DefectStructureComparator(structure, cubic_supercell)
    actual = structure_analyzer.neighboring_atom_indices()
    expected = sorted([25, 15, 16, 23, 54, 47, 46, 56, 62])
    assert actual == expected


def test_actual_files(vasp_files):
    d = vasp_files / "KInO2_Va_O_2"
    calc_results = loadfn(d / "Va_O1_2/calc_results.json")
    perfect_calc_results = loadfn(d / "perfect_calc_results.json")
    structure_comparator = DefectStructureComparator(
        calc_results.structure, perfect_calc_results.structure)
    vac_atom_indices = list(range(192))
    vac_atom_indices.pop(96)
    expected = dict(zip(range(191), vac_atom_indices))
    assert structure_comparator.inserted_indices == []
    assert structure_comparator.removed_indices == [96]
    assert structure_comparator.atom_mapping == expected


@pytest.fixture
def site_diff():
    return SiteDiff(removed=[(1, "Cu", (0.25, 0.74, 0.74))],
                    inserted=[(0,  "Cu", (0.25, 0.5, 0.5))],
                    removed_by_sub=[(0,  "Cu", (0.25, 0.25, 0.25))],
                    inserted_by_sub=[(5, "H", (0.25, 0.25, 0.25))])


def test_site_diff_msonable(site_diff):
    assert_msonable(site_diff)


def test_site_diff_is_no_diff():
    site_diff = SiteDiff(removed=[], inserted=[],
                         removed_by_sub=[], inserted_by_sub=[])
    assert site_diff.is_no_diff is True
    site_diff = SiteDiff(removed=[(1, "H", (0.0, 0.0, 0.0))], inserted=[],
                         removed_by_sub=[], inserted_by_sub=[])
    assert site_diff.is_no_diff is False


def test_site_diff():
    site_diff_vac = SiteDiff(removed=[(1, "H", (0.0, 0.0, 0.0))], inserted=[],
                             removed_by_sub=[], inserted_by_sub=[])
    site_diff_int = SiteDiff(removed=[], inserted=[(1, "H", (0.0, 0.0, 0.0))],
                             removed_by_sub=[], inserted_by_sub=[])
    site_diff_sub = SiteDiff(removed=[], inserted=[],
                             removed_by_sub=[(1, "H", (0.0, 0.0, 0.0))],
                             inserted_by_sub=[(1, "He", (0.0, 0.0, 0.0))])
    assert site_diff_vac.is_complex_defect is False
    assert site_diff_sub.is_complex_defect is False

    assert site_diff_vac.is_vacancy
    assert site_diff_int.is_vacancy is False
    assert site_diff_sub.is_vacancy is False

    assert site_diff_vac.is_interstitial is False
    assert site_diff_int.is_interstitial
    assert site_diff_sub.is_interstitial is False

    assert site_diff_vac.is_substituted is False
    assert site_diff_int.is_substituted is False
    assert site_diff_sub.is_substituted


def test_make_site_diff(structure_comparator, site_diff):
    assert structure_comparator.make_site_diff() == site_diff


def test_make_site_diff_wo_diff():
    s1 = IStructure(Lattice.cubic(10), species=["H"], coords=[[0]*3])
    s2 = s1.copy()
    structure_comparator = DefectStructureComparator(s1, s2)
    site_diff = structure_comparator.make_site_diff()
    assert site_diff == SiteDiff([], [], [], [])
    assert site_diff.is_no_diff

    s1 = IStructure(Lattice.cubic(10), species=["H", "He"], coords=[[0]*3, [0.5]*3])
    s2 = IStructure(Lattice.cubic(10), species=["H"], coords=[[0]*3])
    structure_comparator = DefectStructureComparator(defect_structure=s2,
                                                     perfect_structure=s1)
    site_diff = structure_comparator.make_site_diff()
    assert site_diff == SiteDiff([(1, "He", (0.5, 0.5, 0.5))], [], [], [])
    assert site_diff.is_no_diff is False



# def test_make_defect_type():
#     lattice = Lattice.cubic(10)
#     assert make_defect_type([("")], [], lattice) == DefectType.vacancy
#     assert make_defect_type([], [1], lattice) == DefectType.interstitial
#     assert make_defect_type([3], [1], lattice) == DefectType.interstitial




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pathlib import Path

import yaml
from vise.util.logger import get_logger

""" 
Characteristic data is retrieved from 
McQuarrie and Simon, Phys. Chem. A molecular approach.
"""

logger = get_logger(__name__)

with open(Path(__file__).parent / "molecule_data.yaml", 'r') as f:
    MOLECULE_DATA = yaml.safe_load(f)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from typing import List

import numpy as np
from pydefect.corrections.abstract_correction import Correction
from pydefect.corrections.efnv_correction import \
    ExtendedFnvCorrection, PotentialSite
from tabulate import tabulate


@dataclass
class GkfoCorrection(Correction):
    """
    Potential of gkfo_sites is the difference of potential in the final state
    from that in the initial state.
    And, pc_potential is caused by the additional_charge.
    """
    init_efnv_correction: ExtendedFnvCorrection
    additional_charge: int
    pc_2nd_term: float
    gkfo_sites: List[PotentialSite]
    ave_dielectric_tensor: float
    ave_electronic_dielectric_tensor: float

    def __repr__(self):
        d = [["charge", self.charge],
             ["additional charge", self.additional_charge],
             ["pc 1st term", self.pc_1st_term],
             ["pc 2nd term", self.pc_2nd_term],
             ["alignment 1st term", self.alignment_1st_term],
             ["alignment 2nd term", self.alignment_2nd_term],
             ["alignment 3rd term", self.alignment_3rd_term],
             ["correction energy", self.correction_energy]]
        return tabulate(d, tablefmt='psql')

    @property
    def charge(self):
        return self.init_efnv_correction.charge

    @property
    def defect_region_radius(self):
        return self.init_efnv_correction.defect_region_radius

    @property
    def pc_1st_term(self):
        if self.charge == 0:
            return 0.0
        return (self.init_efnv_correction.point_charge_correction * 2
                / self.charge * self.additional_charge)

    @property
    def sum_pc_correction(self):
        return self.pc_1st_term + self.pc_2nd_term

    @property
    def average_potential_diff_by_addition(self):
        return np.mean([s.diff_pot for s in self.gkfo_sites
                        if s.distance > self.defect_region_radius])

    @property
    def alignment_1st_term(self) -> float:
        return (- self.average_potential_diff_by_addition
                * self.additional_charge)

    @property
    def alignment_2nd_term(self) -> float:
        return (- self.init_efnv_correction.average_potential_diff
                * self.additional_charge)

    @property
    def alignment_3rd_term(self) -> float:
        return - (self.ave_electronic_dielectric_tensor
                  / self.ave_dielectric_tensor
                  * self.charge * self.average_potential_diff_by_addition)

    @property
    def sum_alignment_term(self) -> float:
        return (self.alignment_1st_term + self.alignment_2nd_term
                + self.alignment_3rd_term)

    @property
    def correction_energy(self) -> float:
        return self.sum_pc_correction + self.sum_alignment_term





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from fractions import Fraction
from itertools import cycle
from typing import Optional, List

import plotly.graph_objects as go
from matplotlib import pyplot as plt
from plotly.subplots import make_subplots
from pydefect.analyzer.band_edge_states import BandEdgeOrbitalInfos
from pydefect.defaults import defaults
from vise.util.matplotlib import float_to_int_formatter


class EigenvalueMplSettings:
    def __init__(self,
                 colors: Optional[List[str]] = None,
                 line_width: float = 1.0,
                 band_edge_line_width: float = 1.0,
                 band_edge_line_color: str = "black",
                 band_edge_line_style: str = "-",
                 supercell_band_edge_line_width: float = 0.5,
                 supercell_band_edge_line_color: str = "black",
                 supercell_band_edge_line_style: str = "-.",
                 circle_size: int = 10,
                 tick_label_size: Optional[int] = 10,
                 title_font_size: Optional[int] = 15,
                 label_font_size: Optional[int] = 12):
        self.colors = cycle(colors) if colors else defaults.defect_energy_colors
        self.line_width = line_width
        self.circle_size = circle_size
        self.tick_label_size = tick_label_size
        self.title_font_size = title_font_size
        self.label_font_size = label_font_size

        self.vline = {"linewidth": band_edge_line_width,
                      "color": band_edge_line_color,
                      "linestyle": band_edge_line_style}

        self.svline = {"linewidth": supercell_band_edge_line_width,
                       "color": supercell_band_edge_line_color,
                       "linestyle": supercell_band_edge_line_style}


class EigenvaluePlotter:
    def __init__(self,
                 title: str,
                 band_edge_orb_infos: BandEdgeOrbitalInfos,
                 supercell_vbm: float,
                 supercell_cbm: float,
                 y_range: Optional[List[float]] = None,
                 y_unit: Optional[str] = "eV",
                 ):

        self._title = title
        self._energies_and_occupations = \
            band_edge_orb_infos.energies_and_occupations
        self._kpt_coords = band_edge_orb_infos.kpt_coords
        self._lowest_band_idx = band_edge_orb_infos.lowest_band_index
        self._supercell_vbm = supercell_vbm
        self._supercell_cbm = supercell_cbm
        self._middle = (self._supercell_vbm + self._supercell_cbm) / 2
        self._y_range = y_range
        self._y_unit = y_unit

    def _add_band_idx(self, energy, higher_band_e, lower_band_e):
        if energy < self._middle:
            return higher_band_e - energy > 0.2
        else:
            return energy - lower_band_e > 0.2

    def _x_labels(self, line_break="\n"):
        result = []
        for k in self._kpt_coords:
            x_label = []
            for i in k:
                frac = Fraction(i).limit_denominator(10)
                if frac.numerator == 0:
                    x_label.append("0")
                else:
                    x_label.append(f"{frac.numerator}/{frac.denominator}")
            if x_label == ["0", "0", "0"]:
                result.append("Γ")
            else:
                result.append(line_break.join(x_label))
        return result


class EigenvaluePlotlyPlotter(EigenvaluePlotter):
    def create_figure(self):
        fig = make_subplots(rows=1,
                            cols=2,
                            shared_yaxes=True,
                            horizontal_spacing=0.01,
                            subplot_titles=["up", "down"],
#                            x_title="K-points",
                            )
        for i in fig['layout']['annotations']:
            i['font'] = dict(size=24)

        fig.update_layout(
            title="Eigenvalues",
            yaxis_title="Energy (eV)",
            title_font_size=30,
            font_size=24, width=700, height=700,
            showlegend=False)

        common = {"mode": "markers", "marker_size": 10}

        for spin_idx, eo_by_spin in \
                enumerate(self._energies_and_occupations, 1):
            occupied = [[], [], [], []]
            partially_occupied = [[], [], [], []]
            unoccupied = [[], [], [], []]
            for kpt_idx, eo_by_k_idx in enumerate(eo_by_spin):
                for band_idx, (energy, occup) \
                        in enumerate(eo_by_k_idx, self._lowest_band_idx):
                    if occup > defaults.state_unoccupied_threshold:
                        occupied[0].append(kpt_idx)
                        occupied[1].append(energy)
                        occupied[2].append(band_idx + 1)
                        occupied[3].append(occup)
                    elif occup < defaults.state_occupied_threshold:
                        unoccupied[0].append(kpt_idx)
                        unoccupied[1].append(energy)
                        unoccupied[2].append(band_idx + 1)
                        unoccupied[3].append(occup)
                    else:
                        partially_occupied[0].append(kpt_idx)
                        partially_occupied[1].append(energy)
                        partially_occupied[2].append(band_idx + 1)
                        partially_occupied[3].append(occup)

            fig.add_shape(type="line",
                          x0=-0.5,
                          x1=len(eo_by_spin) - 0.5,
                          y0=self._supercell_vbm,
                          y1=self._supercell_vbm,
                          line=dict(color="MediumPurple", width=4, dash="dot"),
                          row=1, col=spin_idx)
            fig.add_shape(type="line",
                          x0=-0.5,
                          x1=len(eo_by_spin) - 0.5,
                          y0=self._supercell_cbm,
                          y1=self._supercell_cbm,
                          line=dict(color="MediumPurple", width=4, dash="dot"),
                          row=1, col=spin_idx)

            hover_template = '<i>band index</i>: %{text}<br><b>Energy</b>: ' \
                             '%{y:.2f}<br><b>Occupation</b>: %{customdata:.2f}'
            for points, color, name in \
                    zip([occupied, partially_occupied, unoccupied],
                        ["blue", "green", "red"],
                        ["occupied", "partial", "unoccupied"]):
                fig.add_trace(go.Scatter(x=points[0], y=points[1],
                                         text=points[2],
                                         customdata=points[3],
                                         hovertemplate=hover_template,
                                         marker_color=color, name=name,
                                         **common),
                              row=1, col=spin_idx)

                fig.update_xaxes(tickvals=list(range(len(self._kpt_coords))),
                                 ticktext=self._x_labels(line_break="<br>"),
                                 tickfont_size=24,
                                 row=1, col=spin_idx)
        return fig


class EigenvalueMplPlotter(EigenvaluePlotter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._mpl_defaults = kwargs.get("mpl_defaults", EigenvalueMplSettings())
        self.plt = plt

        if len(self._energies_and_occupations) == 2:
            self.fig, self.axs = plt.subplots(nrows=1, ncols=2, sharey='all')
        else:
            self.fig, ax = plt.subplots(nrows=1, ncols=1, sharey='all')
            self.axs = [ax]

    def construct_plot(self):
        self._add_eigenvalues()
        self._add_xticks()
        self._add_band_edges()
        self._set_x_range()
        self._set_labels()
        self._set_title()
        self._set_formatter()
        self.plt.tight_layout()

    def _add_eigenvalues(self):
        for spin_idx, (eo_by_spin, ax) in \
                enumerate(zip(self._energies_and_occupations, self.axs)):
            for kpt_idx, eo_by_k_idx in enumerate(eo_by_spin):
                for band_idx, eo_by_band in enumerate(eo_by_k_idx):
                    energy, occup = eo_by_band
                    color = "r" if occup > 0.9 else "b" if occup < 0.1 else "g"
                    ax.scatter([kpt_idx], energy, marker="o", color=color,
                               s=self._mpl_defaults.circle_size)

                    try:
                        higher_band_e = eo_by_k_idx[band_idx + 1][0]
                        lower_band_e = eo_by_k_idx[band_idx - 1][0]
                    except IndexError:
                        continue

                    if self._add_band_idx(energy, higher_band_e, lower_band_e):
                        ax.annotate(str(band_idx + self._lowest_band_idx + 1),
                                    xy=(kpt_idx + 0.05, energy),
                                    va='center',
                                    fontsize=self._mpl_defaults.tick_label_size)

    def _add_xticks(self):
        for ax in self.axs:
            ax.set_xticks(list(range(len(self._kpt_coords))))
            ax.set_xticklabels(self._x_labels(), size=10)

    def _add_band_edges(self):
        for ax in self.axs:
            ax.axhline(y=self._supercell_vbm, **self._mpl_defaults.svline)
            ax.axhline(y=self._supercell_cbm, **self._mpl_defaults.svline)

    def _set_x_range(self):
        for ax in self.axs:
            ax.set_xlim(-0.5, len(self._kpt_coords) - 0.5)

    def _set_labels(self):
        self.fig.text(0.5, 0, "K-point coords", ha='center',
                      size=self._mpl_defaults.label_font_size)
        self.axs[0].set_ylabel(f"Energy ({self._y_unit})",
                        size=self._mpl_defaults.label_font_size)

    def _set_title(self):
        for ax, spin in zip(self.axs, ["up", "down"]):
            ax.set_title(f"{spin}:")

    def _set_formatter(self):
        self.axs[0].yaxis.set_major_formatter(float_to_int_formatter)
        for ax in self.axs:
            ax.tick_params(labelsize=self._mpl_defaults.tick_label_size)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from functools import reduce
from math import sqrt, pow, ceil
from typing import List, Optional

import numpy as np
from monty.json import MSONable
from numpy import sqrt, dot, pi, exp, cos
from numpy.linalg import norm, inv, det
from pydefect.defaults import defaults
from scipy.special import erfc
from scipy.stats import mstats


def grid_number(lattice_vectors: np.ndarray, max_length: float):
    a = [ceil(max_length / norm(lattice_vectors[i])) for i in range(3)]
    return a[0] * a[1] * a[2]


class Ewald(MSONable):
    """Container class for anisotropic Ewald parameter and related properties

    ewald_param:
        A parameter used for evaluating Ewald sum.
    accuracy:
        product of cutoff radius of G-vector and gaussian FWHM. Increasing this
        value, calculation will be more accurate, but slower.
    """

    def __init__(self,
                 lattice: np.ndarray,
                 dielectric_tensor: np.ndarray,
                 accuracy: float = defaults.ewald_accuracy,
                 ewald_param: Optional[float] = None):
        self.lattice = lattice
        self.rec_lattice = inv(self.lattice).T * 2 * pi
        self.volume = det(self.lattice)
        self.cube_root_vol = pow(self.volume, 1 / 3)

        self.dielectric_tensor = dielectric_tensor
        self.det_epsilon = det(dielectric_tensor)
        self.root_epsilon = sqrt(self.det_epsilon)
        self.epsilon_inv = inv(dielectric_tensor)

        self.accuracy = accuracy
        if ewald_param:
            self.ewald_param = ewald_param
        else:
            l_r = mstats.gmean([norm(v) for v in self.lattice])
            l_g = mstats.gmean([norm(v) for v in self.rec_lattice])
            self.ewald_param = sqrt(l_g / l_r / 2) * self.cube_root_vol / self.root_epsilon

        # Modified Ewald parameter which is the gamma in the Eqs in YK2014.
        self.mod_ewald_param = self.ewald_param / self.cube_root_vol * self.root_epsilon
        # 2nd term in Eq.(14) in YK2014, caused by finite gaussian charge.
        self.diff_pot = -0.25 / self.volume / self.mod_ewald_param ** 2

    def atomic_site_potential(self, rel_coord):
        real_part = self.ewald_real(include_self=True, shift=rel_coord)
        rec_part = self.ewald_rec(rel_coord)
        return real_part + rec_part + self.diff_pot

    @property
    def lattice_energy(self):
        real_part = self.ewald_real(include_self=False, shift=[0, 0, 0])
        rec_part = self.ewald_rec([0, 0, 0])
        return (real_part + rec_part + self.diff_pot + self.self_pot) / 2

    @property
    def self_pot(self):
        return - self.mod_ewald_param / (2.0 * pi * sqrt(pi * self.det_epsilon))

    def ewald_real(self, include_self: bool, shift: List[float]) -> float:
        summed = 0
        for r in self.r_lattice_set(include_self, shift):
            root_r_inv_epsilon_r = sqrt(reduce(dot, [r.T, self.epsilon_inv, r]))
            summed += erfc(self.mod_ewald_param * root_r_inv_epsilon_r) / root_r_inv_epsilon_r
        return summed / (4 * pi * self.root_epsilon)

    def ewald_rec(self, coord) -> float:
        cart_coord = dot(coord, self.lattice)
        summed = 0
        for g in self.g_lattice_set():
            g_epsilon_g = reduce(dot, [g.T, self.dielectric_tensor, g])
            summed += (exp(- g_epsilon_g / 4 / self.mod_ewald_param ** 2)
                       / g_epsilon_g * cos(dot(g, cart_coord)))
        return summed / self.volume

    def r_lattice_set(self,
                      include_self: bool = True,
                      shift: List[float] = None) -> np.ndarray:
        xyz = self.xyz(self.r_vector_nums, shift)
        if not include_self:
            xyz = np.delete(xyz, int((len(xyz) - 1) / 2), 0)
        return dot(xyz, self.lattice)

    def g_lattice_set(self) -> np.ndarray:
        xyz = self.xyz(self.g_vector_nums)
        xyz = np.delete(xyz, int((len(xyz) - 1) / 2), 0)
        return dot(xyz, self.rec_lattice)

    @staticmethod
    def xyz(nums: List[int], shift: List[float] = None) -> np.ndarray:
        shift = shift or [0, 0, 0]
        x = np.arange(-nums[0], nums[0] + 1, 1) - shift[0]
        y = np.arange(-nums[1], nums[1] + 1, 1) - shift[1]
        z = np.arange(-nums[2], nums[2] + 1, 1) - shift[2]
        return np.array(np.meshgrid(x, y, z)).T.reshape(-1, 3)

    @property
    def r_vector_nums(self) -> List[int]:
        max_length = self.accuracy / self.mod_ewald_param
        return [ceil(max_length / norm(self.lattice[i])) for i in range(3)]

    @property
    def g_vector_nums(self) -> List[int]:
        max_length = 2 * self.mod_ewald_param * self.accuracy
        return [ceil(max_length / norm(self.rec_lattice[i])) for i in range(3)]




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest
from pydefect.corrections.abstract_correction import Correction
from pydefect.corrections.efnv_correction import PotentialSite
from pydefect.corrections.gkfo_correction import GkfoCorrection


@pytest.fixture
def gkfo_correction(efnv_correction):
    s1 = PotentialSite(
        specie="H", distance=1.999, potential=10, pc_potential=None)
    s2 = PotentialSite(
        specie="He", distance=2.0001, potential=15, pc_potential=2)
    s3 = PotentialSite(
        specie="He", distance=3.0, potential=20, pc_potential=3)

    return GkfoCorrection(init_efnv_correction=efnv_correction,
                          additional_charge=1,
                          pc_2nd_term=10.0,
                          gkfo_sites=[s1, s2, s3],
                          ave_dielectric_tensor=4.0,
                          ave_electronic_dielectric_tensor=2.0)


def test_is_correction(gkfo_correction):
    assert isinstance(gkfo_correction, Correction)


def test_repr(gkfo_correction):
    expected = """+--------------------+-------+
| charge             |  10   |
| additional charge  |   1   |
| pc 1st term        |   0.2 |
| pc 2nd term        |  10   |
| alignment 1st term | -15   |
| alignment 2nd term |  -1.5 |
| alignment 3rd term | -75   |
| correction energy  | -81.3 |
+--------------------+-------+"""
    assert repr(gkfo_correction) == expected


def test_gkfo_correction_charge(gkfo_correction):
    assert gkfo_correction.charge == 10


def test_gkfo_correction_defect_region_radius(gkfo_correction):
    assert gkfo_correction.defect_region_radius == 2.0


def test_gkfo_correction_pc_1st_term(gkfo_correction):
    assert gkfo_correction.pc_1st_term == 1.0 * 2 / 10 * 1


def test_gkfo_correction_sum_pc_correction(gkfo_correction):
    assert gkfo_correction.sum_pc_correction == 1.0 * 2 / 10 * 1 + 10.0


def test_extended_fnv_correction_alignment_correction(gkfo_correction):
    assert gkfo_correction.average_potential_diff_by_addition == ((15 - 2) + (20 - 3)) / 2
    assert gkfo_correction.alignment_1st_term == - ((15 - 2) + (20 - 3)) / 2
    assert gkfo_correction.alignment_2nd_term == - ((1.5 - 0.2) + (2 - 0.3)) / 2
    assert gkfo_correction.alignment_3rd_term == - ((15 - 2) + (20 - 3)) / 2 * 10 * 2 / 4
    assert gkfo_correction.sum_alignment_term == - 15 - 1.5 - 75


def test_extended_fnv_correction_correction_energy(gkfo_correction):
    assert gkfo_correction.correction_energy == 0.2 + 10.0 - 15 - 1.5 - 75





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from itertools import cycle
from typing import List, Optional, Tuple

import plotly.graph_objects as go
from adjustText import adjust_text
from labellines import labelLines
from matplotlib import pyplot as plt
from pydefect.analyzer.defect_energy import DefectEnergySummary
from pydefect.defaults import defaults
from vise.util.matplotlib import float_to_int_formatter


class DefectEnergiesMplSettings:
    def __init__(self,
                 colors: Optional[List[str]] = None,
                 line_width: float = 1.0,
                 vline_width: float = 0.5,
                 vline_color: str = "black",
                 vline_style: str = "-.",
                 vline_alpha: float = 0.7,
                 circle_size: int = 15,
                 tick_label_size: Optional[int] = 12,
                 title_font_size: Optional[int] = 15,
                 label_font_size: Optional[int] = 15,
                 defect_name_size: Optional[int] = 12,
                 charge_size: Optional[int] = 12):
        self.colors = cycle(colors) if colors else defaults.defect_energy_colors
        self.line_width = line_width
        self.circle_size = circle_size
        self.tick_label_size = tick_label_size
        self.title_font_size = title_font_size
        self.label_font_size = label_font_size
        self.defect_name_size = defect_name_size
        self.charge_size = charge_size

        self.vline = {"linewidth": vline_width,
                      "color":     vline_color,
                      "linestyle": vline_style,
                      "alpha": vline_alpha}


class DefectEnergyPlotter:
    def __init__(self,
                 defect_energy_summary: DefectEnergySummary,
                 chem_pot_label: "str",
                 allow_shallow: bool,
                 with_corrections: bool,
                 name_style: Optional[str],
                 x_range: Optional[Tuple[float, float]] = None,
                 y_range: Optional[Tuple[float, float]] = None,
                 vline_threshold: float = 0.02,
                 x_unit: Optional[str] = "eV",
                 y_unit: Optional[str] = "eV",
                 **plot_settings):

        self._title = defect_energy_summary.latexified_title
        self._supercell_vbm = defect_energy_summary.supercell_vbm
        self._supercell_cbm = defect_energy_summary.supercell_cbm
        self._x_range = x_range or (0, defect_energy_summary.cbm)
        defect_energy_summary.e_min = self._x_range[0]
        defect_energy_summary.e_max = self._x_range[1]
        charge_energies = defect_energy_summary.charge_energies(
            chem_pot_label, allow_shallow, with_corrections, self._x_range,
            name_style)
        self._cross_points = charge_energies.cross_point_dicts
        self._y_range = y_range or charge_energies.energy_range(space=0.2)
        self._vline_threshold = vline_threshold
        self._x_unit = x_unit
        self._y_unit = y_unit


class DefectEnergyPlotlyPlotter(DefectEnergyPlotter):
    def __init__(self, **kwargs):
        super().__init__(name_style="plotly", **kwargs)

    def create_figure(self):
        fig = go.Figure()
        fig.update_layout(
            title=f"Defect formation energy in {self._title}",
            xaxis_title=f"Fermi level ({self._x_unit})",
            yaxis_title=f"Energy ({self._y_unit})",
            title_font_size=30,
            font_size=24,
            width=900, height=700)

        for name, cp in self._cross_points.items():
            xs, ys = cp.t_all_sorted_points
            fig.add_trace(go.Scatter(x=xs, y=ys, name=name,
                                     text=cp.charge_list,
                                     hovertemplate=
                                     f'{name}<br>' +
                                     'Charges %{text}<br>' +
                                     'Energy: %{y:.2f}',
                                     line_width=3,
                                     marker_size=15))

        fig["layout"]["xaxis"]["range"] = [self._x_range[0] - 0.1,
                                           self._x_range[1] + 0.1]
        fig["layout"]["yaxis"]["range"] = self._y_range

        kwargs = dict(y=self._y_range + [None] + self._y_range,
                      line=dict(width=2, dash="dash"),
                      line_color="black")
        name = "unitcell"
        fig.add_trace(go.Scatter(
            x=[self._x_range[0]]*2 + [None] + [self._x_range[1]]*2,
            name=name, showlegend=True, **kwargs))

        x = [None] * 5
        if self._supercell_vbm > self._vline_threshold:
            x[0], x[1] = self._supercell_vbm, self._supercell_vbm
        if self._supercell_cbm < self._x_range[1] - self._vline_threshold:
            x[3], x[4] = self._supercell_cbm, self._supercell_cbm

        if len(set(x)) > 1:
            kwargs["name"] = "supercell"
            kwargs["line_color"] = "blue"
            kwargs["line"]["dash"] = "dot"
            fig.add_trace(go.Scatter(x=x, showlegend=True, **kwargs))

        return fig


class DefectEnergyMplPlotter(DefectEnergyPlotter):
    def __init__(self,
                 label_line: bool = True,
                 add_charges: bool = True,
                 **kwargs):
        super().__init__(name_style="mpl", **kwargs)
        self._mpl_defaults = \
            kwargs.get("mpl_defaults", DefectEnergiesMplSettings())
        self._label_line = label_line
        self._add_charges = add_charges
        self.plt = plt
        self._texts = []

    def construct_plot(self):
        self._add_energies()
        self._add_band_edges()
        self._set_x_range()
        self._set_y_range()
        self._set_labels()
        self._set_title()
        self._set_formatter()

        if self._add_charges:
            adjust_text(self._texts, force_points=(1.0, 2.5))

        if self._label_line:
            labelLines(plt.gca().get_lines(),
                       align=False,
                       fontsize=self._mpl_defaults.defect_name_size)
        else:
            ax = self.plt.gca()
            ax.legend(bbox_to_anchor=(1, 0.5), loc='center left')

        self.plt.tight_layout()

    def _add_energies(self):
        for name, cp in self._cross_points.items():
            color = next(self._mpl_defaults.colors)
            self.plt.plot(*cp.t_all_sorted_points, color=color,
                          linewidth=self._mpl_defaults.line_width,
                          label=name)
            if cp.t_inner_cross_points:
                self.plt.scatter(*cp.t_inner_cross_points, marker="o",
                                 color=color, s=self._mpl_defaults.circle_size)
            if self._add_charges:
                self._texts.extend(
                    [self.plt.text(x, y, charge, color=color,
                                   fontsize=self._mpl_defaults.charge_size)
                     for charge, (x, y) in cp.annotated_charge_positions.items()])

    def _set_x_range(self):
        self.plt.xlim(self._x_range)

    def _set_y_range(self):
        if self._y_range:
            self.plt.ylim(self._y_range[0], self._y_range[1])

    def _set_labels(self):
        self.plt.xlabel(f"Fermi level ({self._x_unit})",
                        size=self._mpl_defaults.label_font_size)
        self.plt.ylabel(f"Energy ({self._y_unit})",
                        size=self._mpl_defaults.label_font_size)

    def _set_title(self):
        self.plt.title(self._title, size=self._mpl_defaults.title_font_size)

    def _set_formatter(self):
        axis = self.plt.gca()
        axis.yaxis.set_major_formatter(float_to_int_formatter)
        axis.tick_params(labelsize=self._mpl_defaults.tick_label_size)

    def _add_band_edges(self):
        if self._supercell_vbm > self._vline_threshold:
            self.plt.axvline(x=self._supercell_vbm,
                             **self._mpl_defaults.vline)
            plt.text(self._supercell_vbm, self._y_range[1], 'supercell VBM',
                     size=8, ha='center', va='center', rotation='vertical',
                     backgroundcolor='white')
        if self._supercell_cbm < self._x_range[1] - self._vline_threshold:
            self.plt.axvline(x=self._supercell_cbm,
                             **self._mpl_defaults.vline)
            plt.text(self._supercell_cbm, self._y_range[1], 'supercell',
                     size=8, ha='center', va='center', rotation='vertical',
                     backgroundcolor='white')



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from typing import List, Optional

import numpy as np
from pydefect.analyzer.defect_charge_info import DefectChargeInfo, \
    AveChargeDensityDist
from pydefect.analyzer.grids import Grids
from pydefect.cli.vasp.make_efnv_correction import calc_max_sphere_radius
from pymatgen.io.vasp import Chgcar
from vise.analyzer.vasp.handle_volumetric_data import make_spin_charges


def center_1d_periodic_quantity(grid_points: List[float]):
    num_grid_pts = len(grid_points)
    moments = []
    for i in range(num_grid_pts):
        moment = 0.0
        if num_grid_pts % 2 == 0:
            for j in range(1, int(num_grid_pts / 2 + 1)):
                moment += grid_points[i - j] * j
            for j in range(1, int(num_grid_pts / 2)):
                k = (i + j - num_grid_pts) % num_grid_pts
                moment += grid_points[k] * j
        else:
            a = int(np.floor(num_grid_pts / 2))
            for j in range(1, a + 1):
                moment += grid_points[i - j] * j
            for j in range(1, a + 1):
                k = (i + j - num_grid_pts) % num_grid_pts
                moment += grid_points[k] * j

        moments.append(moment)

    return np.nanargmin(moments)


def make_charge_dist(parchg: Chgcar, grids: Grids, distance_bins: np.ndarray):
    assert parchg.structure.lattice == grids.lattice
    assert parchg.dim == grids.dim

    spin_charges = make_spin_charges(parchg)

    dists = []
    defect_center_idxs = []
    for chg in spin_charges:
        center = [center_1d_periodic_quantity(
            chg.get_average_along_axis(i)) for i in [0, 1, 2]]
        defect_center_idxs.append(np.array(center))
        data = chg.data["total"]
        dists.append(grids.spherical_dist(data, center, distance_bins))

    result = []
    for c, d in zip(defect_center_idxs, dists):
        result.append(AveChargeDensityDist(tuple(c / grids.dim), d))

    return result


def make_defect_charge_info(parchgs: List[Chgcar],
                            band_idxs: List[int],
                            bin_interval: float,
                            grids: Grids = None):
    if grids is None:
        grids = Grids.from_chgcar(parchgs[0])

    radius = calc_max_sphere_radius(parchgs[0].structure.lattice.matrix)
    num_bins = int(np.ceil(radius / bin_interval))
    distance_bins = np.array([bin_interval * i for i in range(num_bins)] + [radius])
    ave_charge_density = 1.0 / parchgs[0].structure.volume
    charge_dists = []
    for parchg in parchgs:
        charge_dists.append(make_charge_dist(parchg, grids, distance_bins))
    return DefectChargeInfo(distance_bins.tolist(), band_idxs, charge_dists,
                            ave_charge_density)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pydefect.analyzer.make_defect_structure_info import \
    MakeDefectStructureInfo
from pydefect.cli.make_defect_vesta_file import fold_coords_in_structure

from pymatgen.core import Structure, Lattice
from vise.tests.helpers.assertion import assert_dataclass_almost_equal


def test_fold_coords():
    actual = Structure(Lattice.monoclinic(a=1, b=1, c=1, beta=20),
                       species=["H"],
                       coords=[[0.5, 0, 0.49]])
    fold_coords_in_structure(actual, [0, 0, 0])
    expected = Structure(Lattice.monoclinic(a=1, b=1, c=1, beta=20),
                         species=["H"],
                         coords=[[-0.5, 0, 0.49]])
    assert actual == expected


def test_make_defect_structure_info(structures, def_str_info):
    perf, initial, final = structures
    info = MakeDefectStructureInfo(perf, initial, final,  dist_tol=0.2,
                                   symprec=0.1,
                                   neighbor_cutoff_factor=1.2)
    actual = info.defect_structure_info
    assert_dataclass_almost_equal(actual, def_str_info, check_is_subclass=True)


def test_make_defect_structure_info2():
    perf = Structure(Lattice.cubic(1), ["H"], [[0., 0., 0.]])
    init = Structure(Lattice.cubic(1), ["H"]*2, [[0., 0., 0.], [0.5, 0.0, 0.0]])
    fin = Structure(Lattice.cubic(1), ["H"]*2, [[0., 0., 0.], [0.4, 0.0, 0.0]])

    info = MakeDefectStructureInfo(perf, init, fin,  dist_tol=0.05,
                                   symprec=0.1,
                                   neighbor_cutoff_factor=1.2)
    actual = info.defect_structure_info
    print(actual)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from argparse import Namespace
from pathlib import Path

from pydefect.analyzer.band_edge_states import PerfectBandEdgeState
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.defect_energy import DefectEnergySummary
from pydefect.cli.main import parse_args_main
from pydefect.input_maker.supercell_info import SupercellInfo


def test_make_standard_and_relative_energies_options():
    parsed_args = parse_args_main(["sre"])
    expected = Namespace(
        composition_energies_yaml="composition_energies.yaml",
        func=parsed_args.func)
    assert parsed_args == expected


def test_make_cpd_and_vertices_options():
    parsed_args = parse_args_main(["cv",
                                   "-t", "MgO",
                                   "-e", "Mg", "O", "Al"])
    expected = Namespace(
        rel_energy_yaml="relative_energies.yaml",
        target="MgO",
        elements=["Mg", "O", "Al"],
        func=parsed_args.func)
    assert parsed_args == expected


def test_plot_cpd_options(mocker):
    mock_loadfn = mocker.patch("pydefect.cli.main.loadfn")
    parsed_args = parse_args_main(["pc"])
    expected = Namespace(
        chem_pot_diag=mock_loadfn.return_value,
        func=parsed_args.func)
    assert parsed_args == expected


def test_make_supercell_wo_options(mocker):
    mock = mocker.patch("pydefect.cli.main.IStructure")
    parsed_args = parse_args_main(["s", "-p", "POSCAR-tmp"])
    # func is a pointer so need to point the same address.
    expected = Namespace(
        unitcell=mock.from_file.return_value,
        matrix=None,
        min_num_atoms=50,
        max_num_atoms=300,
        func=parsed_args.func)
    assert parsed_args == expected


def test_make_supercell_w_options(mocker):
    mock = mocker.patch("pydefect.cli.main.IStructure")
    parsed_args = parse_args_main(["s",
                                   "-p", "POSCAR-tmp",
                                   "--matrix", "1", "2", "3",
                                   "--min_atoms", "1000",
                                   "--max_atoms", "2000"])
    # func is a pointer so need to point the same address.
    expected = Namespace(
        unitcell=mock.from_file.return_value,
        matrix=[1, 2, 3],
        min_num_atoms=1000,
        max_num_atoms=2000,
        func=parsed_args.func)
    assert parsed_args == expected
    mock.from_file.assert_called_once_with("POSCAR-tmp")


def test_append_interstitial(mocker):

    mock_loadfn = mocker.patch("pydefect.cli.main.loadfn")
    mock_supercell_info = mocker.Mock(spec=SupercellInfo, autospec=True)
    mock_loadfn.return_value = mock_supercell_info
    mock_structure = mocker.patch("pydefect.cli.main.Structure")
    parsed_args = parse_args_main(["ai",
                                   "-s", "supercell_info.json",
                                   "-p", "POSCAR",
                                   "-c", "0.1", "0.2", "0.3",
                                   "-i", "info"])
    expected = Namespace(
        supercell_info=mock_supercell_info,
        base_structure=mock_structure.from_file.return_value,
        frac_coords=[0.1, 0.2, 0.3],
        info="info",
        func=parsed_args.func)
    assert parsed_args == expected


def test_pop_interstitial(mocker):

    mock_loadfn = mocker.patch("pydefect.cli.main.loadfn")
    mock_supercell_info = mocker.Mock(spec=SupercellInfo, autospec=True)
    mock_loadfn.return_value = mock_supercell_info
    parsed_args = parse_args_main(["pi",
                                   "-s", "supercell_info.json",
                                   "-i", "1000",
                                   "--pop_all"])
    expected = Namespace(
        supercell_info=mock_supercell_info,
        index=1000,
        pop_all=True,
        func=parsed_args.func)
    assert parsed_args == expected


def test_defect_set_wo_options():
    parsed_args = parse_args_main(["ds"])
    expected = Namespace(
        oxi_states=None,
        dopants=None,
        keywords=None,
        func=parsed_args.func)
    assert parsed_args == expected


def test_defect_set_w_options():
    parsed_args = parse_args_main(["ds",
                                   "-o", "He", "1",
                                   "-d", "Li",
                                   "-k", "Li_H1", "Va_H1_0"])
    expected = Namespace(
        oxi_states=["He", 1],
        dopants=["Li"],
        keywords=["Li_H1", "Va_H1_0"],
        func=parsed_args.func)
    assert parsed_args == expected


def test_defect_structure_info(mocker):
    mock_supercell_info = mocker.Mock(spec=SupercellInfo, autospec=True)

    def side_effect(filename):
        if filename == "supercell_info.json":
            return mock_supercell_info
        else:
            raise ValueError

    mock_loadfn = mocker.patch("pydefect.cli.main.loadfn",
                               side_effect=side_effect)
    parsed_args = parse_args_main(["dsi",
                                   "-s", "supercell_info.json",
                                   "-d", "Va_O1_0",
                                   "-nccr",
                                   "-dt", "1.0",
                                   "--symprec", "2.0"])
    expected = Namespace(
        supercell_info=mock_supercell_info,
        check_calc_results=False,
        dirs=[Path("Va_O1_0")],
        dist_tolerance=1.0,
        symprec=2.0,
        verbose=False,
        func=parsed_args.func)
    assert parsed_args == expected


def test_efnv_correction(mocker):
    mock_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_unitcell = mocker.patch("pydefect.cli.main.Unitcell")

    def side_effect(filename):
        if filename == "perfect/calc_results.json":
            return mock_calc_results
        else:
            raise ValueError

    mocker.patch("pydefect.cli.main.loadfn", side_effect=side_effect)
    parsed_args = parse_args_main(["efnv",
                                   "-d", "Va_O1_0", "Va_O1_1",
                                   "-nccr",
                                   "-pcr", "perfect/calc_results.json",
                                   "-u", "unitcell.json"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        check_calc_results=False,
        perfect_calc_results=mock_calc_results,
        unitcell=mock_unitcell.from_yaml.return_value,
        verbose=False,
        func=parsed_args.func)
    assert parsed_args == expected


def test_band_edge_states(mocker):
    mock = mocker.patch("pydefect.cli.main.loadfn")
    parsed_args = parse_args_main([
        "bes",
        "-d", "Va_O1_0", "Va_O1_1",
        "-nccr",
        "-p", "perfect/perfect_band_edge_state.json"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        check_calc_results=False,
        p_state=mock.return_value,
        verbose=False,
        func=parsed_args.func)
    assert parsed_args == expected
    mock.assert_any_call("perfect/perfect_band_edge_state.json")


def test_defect_energy_infos(mocker):
    mock_unitcell = mocker.patch("pydefect.cli.main.Unitcell")
    mock_loadfn = mocker.patch("pydefect.cli.main.loadfn")
    mock_std_energy = mocker.patch("pydefect.cli.main.StandardEnergies")
    parsed_args = parse_args_main([
        "dei",
        "-d", "Va_O1_0", "Va_O1_1",
        "-nccr",
        "-u", "unitcell.json",
        "-p", "perfect/perfect_band_edge_state.json",
        "-s", "standard_energies.yaml"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        check_calc_results=False,
        unitcell=mock_unitcell.from_yaml.return_value,
        perfect_calc_results=mock_loadfn.return_value,
        std_energies=mock_std_energy.from_yaml.return_value,
        verbose=False,
        func=parsed_args.func)
    assert parsed_args == expected
    mock_loadfn.assert_called_once_with("perfect/perfect_band_edge_state.json")
    mock_std_energy.from_yaml.assert_called_once_with("standard_energies.yaml")


def test_defect_energy_summary(mocker):
    mock_pbes = mocker.Mock(spec=PerfectBandEdgeState, autospec=True)
    mock_unitcell = mocker.patch("pydefect.cli.main.Unitcell")

    def side_effect(filename):
        if filename == "perfect/calc_results.json":
            return mock_pbes
        else:
            print(filename)
            raise ValueError

    mocker.patch("pydefect.cli.main.loadfn", side_effect=side_effect)
    parsed_args = parse_args_main(["des",
                                   "-d", "Va_O1_0", "Va_O1_1",
                                   "-p", "perfect/calc_results.json",
                                   "-u", "unitcell.json",
                                   "-t", "target_vertices.yaml"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        p_state=mock_pbes,
        unitcell=mock_unitcell.from_yaml.return_value,
        target_vertices_yaml="target_vertices.yaml",
        verbose=False,
        func=parsed_args.func)
    assert parsed_args == expected


def test_plot_defect_formation_energy(mocker):
    mock_summary = mocker.Mock(spec=DefectEnergySummary, autospec=True)

    def side_effect(filename):
        if filename == "defect_energy_summary.json":
            return mock_summary
        else:
            raise ValueError

    mocker.patch("pydefect.cli.main.loadfn", side_effect=side_effect)
    parsed_args = parse_args_main([
        "pe", "-d", "defect_energy_summary.json", "-l", "A", "-y", "-5", "5",
        "--allow_shallow", "--no_corrections", "--no_label_line",
        "--no_add_charges"])

    expected = Namespace(
        defect_energy_summary=mock_summary,
        label="A",
        y_range=[-5, 5],
        allow_shallow=True,
        with_corrections=False,
        label_line=False,
        add_charges=False,
        func=parsed_args.func)
    assert parsed_args == expected




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import numpy as np
import pytest
from pydefect.input_maker.supercell import (
    Supercell, Supercells, TetragonalSupercells, RhombohedralSupercells)
from pydefect.util.error_classes import SupercellError
from pymatgen.core import Element


def test_supercell(simple_cubic, simple_cubic_2x1x1):
    matrix = [[2, 0, 0], [0, 1, 0], [0, 0, 1]]
    supercell = Supercell(input_structure=simple_cubic, matrix=matrix)
    assert supercell.structure == simple_cubic_2x1x1

    average = (2 + 1 + 1) / 3
    expected = (abs(2 - average) + abs(1 - average) * 2) / 3 / average
    assert supercell.isotropy == expected


def test_supercell_species_order(complex_monoclinic):
    matrix = [[2, 0, 0], [0, 1, 0], [0, 0, 1]]
    supercell = Supercell(input_structure=complex_monoclinic, matrix=matrix)
    actual = [e.specie for e in supercell.structure]
    expected = [Element.H] * 2 + [Element.He] * 8
    assert actual == expected


def test_supercell_average_angle(monoclinic):
    matrix = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    supercell = Supercell(input_structure=monoclinic, matrix=matrix)
    assert supercell.average_angle == (90 + 90 + 100) / 3


def test_supercells(simple_cubic):
    supercells = Supercells(input_structure=simple_cubic,
                            min_num_atoms=64,
                            max_num_atoms=100)
    expected = simple_cubic * [[4, 0, 0], [0, 4, 0], [0, 0, 4]]
    assert isinstance(supercells.most_isotropic_supercell, Supercell)
    assert supercells.most_isotropic_supercell.structure == expected


def test_supercells_raise_no_supercell_error(simple_cubic):
    supercells = Supercells(input_structure=simple_cubic,
                            min_num_atoms=10,
                            max_num_atoms=10)
    with pytest.raises(SupercellError):
        print(supercells.most_isotropic_supercell)


def test_rhombohedral_supercells(rhombohedral):
    supercells = RhombohedralSupercells(input_structure=rhombohedral,
                                        min_num_atoms=2,
                                        max_num_atoms=4)
    actual = supercells.most_isotropic_supercell.lattice.angles[0]
    assert actual == 74.85849218561553


def test_tetragonal_supercells(elongated_tetragonal):
    supercells = TetragonalSupercells(input_structure=elongated_tetragonal,
                                      min_num_atoms=2,
                                      max_num_atoms=300)
    actual = supercells.most_isotropic_supercell.lattice.lengths
    expected = (4.242640687119285, 4.242640687119285, 4.242640687119286)
    np.testing.assert_array_almost_equal(actual, expected)

    actual = supercells.most_isotropic_supercell.matrix
    expected = np.array([[3, 3, 0], [-3, 3, 0], [0, 0, 1]])
    np.testing.assert_array_equal(actual, expected)


def test_matrix_from_x_y():
    actual = TetragonalSupercells.matrix_from_x_y(2, 0)
    expected = np.array([[2, 0], [0, 2]])
    np.testing.assert_array_equal(actual, expected)

    actual = TetragonalSupercells.matrix_from_x_y(2, 1)
    expected = np.array([[2, 2], [-2, 2]])
    np.testing.assert_array_equal(actual, expected)

    actual = TetragonalSupercells.matrix_from_x_y(1, 2)
    expected = np.array([[0, 2], [-2, 0]])  # rotate the axis
    np.testing.assert_array_equal(actual, expected)


def test_next_x_y_combination():
    assert TetragonalSupercells.next_x_y_combination(24) == (5, 0)
    assert TetragonalSupercells.next_x_y_combination(25) in [(2, 3), (1, 5)]







# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pydefect.cli.vasp.make_calc_results import make_calc_results_from_vasp
from pymatgen.core import Structure
from pymatgen.io.vasp import Vasprun, Outcar


def test_make_calc_results_from_vasp_results(vasp_files):
    vasprun = Vasprun(vasp_files / "MgO_conv_Va_O_0" / "vasprun.xml")
    outcar = Outcar(vasp_files / "MgO_conv_Va_O_0" / "OUTCAR")
    results = make_calc_results_from_vasp(vasprun, outcar)

    expected_structure = \
        Structure.from_file(vasp_files / "MgO_conv_Va_O_0" / "CONTCAR")
    assert results.structure == expected_structure
    assert results.energy == -34.91084360
    assert results.magnetization == 1.03e-05
    assert results.potentials ==\
           [35.9483, 36.066, 35.948, 35.9478, 69.799, 69.7994, 69.7995]



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pathlib import Path

from monty.serialization import loadfn
from vise.util.logger import get_logger

logger = get_logger(__name__)

electroneg_data = loadfn(Path(__file__).parent / "electronegativity.yaml")
oxi_state_data = loadfn(Path(__file__).parent / "oxidation_state.yaml")


def electronegativity(element):
    if element not in electroneg_data:
        logger.info(f"Electronegativity of {element} is set to None.")
        electroneg_data[element] = None
    return electroneg_data[element]


def oxidation_state(element):
    if element not in oxi_state_data:
        logger.info(f"Oxidation state of {element} is set to 0.")
        oxi_state_data[element] = 0
    return oxi_state_data[element]




# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

import argparse
import sys
import warnings

from monty.serialization import loadfn
from pydefect.cli.main import description, epilog, add_sub_parser
from pydefect.cli.main_util_functions import make_gkfo_correction_from_vasp, \
    composition_energies_from_mp, add_interstitials_from_local_extrema, \
    make_defect_vesta_file
from pydefect.defaults import defaults
from pymatgen.io.vasp.inputs import UnknownPotcarWarning

warnings.simplefilter('ignore', UnknownPotcarWarning)


def parse_args_main_util(args):
    parser = argparse.ArgumentParser(epilog=epilog,
                                     description=description + """                            
    This command provide some utilities related to the VASP calculations""")

    subparsers = parser.add_subparsers()
    dirs_parser = add_sub_parser(argparse, name="dirs")
    unitcell_parser = add_sub_parser(argparse, name="unitcell")
    si_parser = add_sub_parser(argparse, name="supercell_info")
    no_calc_results = add_sub_parser(argparse, name="no_calc_results_check")

    # -- composition energies from mp ------------------------------------------
    parser_comp_es_from_mp = subparsers.add_parser(
        name="composition_energies_from_mp",
        description="Retrieve composition energies from Materials Project.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['cefm'])

    parser_comp_es_from_mp.add_argument(
        "-e", "--elements", type=str, nargs="+",
        help="Element names considered in chemical potential diagram. Used for "
             "creating the diagram from mp data with atom energy alignment.")
    parser_comp_es_from_mp.add_argument(
        "-a", "--atom_energy_yaml", type=str,
        help="Yaml file storing atom energies for energy alignment.")

    parser_comp_es_from_mp.set_defaults(func=composition_energies_from_mp)

    # -- add interstitials from local extrema ----------------------------------
    parser_ai = subparsers.add_parser(
        name="add_interstitials_from_local_extrema",
        description="Add interstitials via volumetric_data_local_extrema.json "
                    "file.",
        parents=[si_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['ai'])
    parser_ai.add_argument(
        "--local_extrema", required=True, type=loadfn,
        help="volumetric_data_local_extrema.json file name.")
    parser_ai.add_argument(
        "-i", "--indices", required=True, type=int, nargs="+",
        help="Indices starting from 1 to be added to SupercellInfo.")
    parser_ai.set_defaults(func=add_interstitials_from_local_extrema)

    # -- make defect vesta file ------------------------------------------------
    parser_dvf = subparsers.add_parser(
        name="defect_vesta_file",
        description="Generate VESTA file for visualizing defect structure.",
        parents=[dirs_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['dvf'])
    parser_dvf.add_argument(
        "--cutoff", type=float, default=defaults.show_structure_cutoff,
        help="Cutoff radius in which atoms are shown in VESTA files.")
    parser_dvf.add_argument(
        "--min_displace_w_arrows", type=float, default=0.1,
        help="Minimum displacement that shows arrows to visualize "
             "displacements.")
    parser_dvf.add_argument(
        "--arrow_factor", type=float, default=3.0,
        help="Factor to enhance the displacements.")
    parser_dvf.add_argument(
        "--title", type=str, help="Title to be shown in VESTA files.")

    parser_dvf.set_defaults(func=make_defect_vesta_file)

    # -- gkfo correction -------------------------------------------------------
    parser_gkfo = subparsers.add_parser(
        name="gkfo",
        description="Generate GKFO correction files.",
        parents=[unitcell_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['gkfo'])

    parser_gkfo.add_argument(
        "-iefnv", "--initial_efnv_correction", required=True, type=loadfn,
        help="Path to the initial efnv correction.json file.")
    parser_gkfo.add_argument(
        "-icr", "--initial_calc_results", required=True, type=loadfn,
        help="Path to the initial calc_results.json file.")
    parser_gkfo.add_argument(
        "-fcr", "--final_calc_results", required=True, type=loadfn,
        help="Path to the final calc_results.json file.")
    parser_gkfo.add_argument(
        "-cd", "--charge_diff", required=True, type=int,
        help="Charge difference of final state from initial state.")

    parser_gkfo.set_defaults(func=make_gkfo_correction_from_vasp)
    # ------------------------------------------------------------------------
    return parser.parse_args(args)


def main():
    args = parse_args_main_util(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

import numpy as np
from numpy.linalg import inv
from pydefect.analyzer.defect_structure_info import logger
from pydefect.defaults import defaults
from pymatgen.core import Structure
from vise.util.structure_symmetrizer import StructureSymmetrizer


def refine_defect_structure(structure: Structure,
                            anchor_atom_index: int = None,
                            anchor_atom_coords: np.ndarray = None):
    symmetrizer = StructureSymmetrizer(structure,
                                       defaults.symmetry_length_tolerance,
                                       defaults.symmetry_angle_tolerance)
    result = structure.copy()
    spglib_data = symmetrizer.spglib_sym_data

    origin_shift = spglib_data["origin_shift"]
    inv_trans_mat = inv(spglib_data["transformation_matrix"])
    coords = spglib_data["std_positions"]
    # This transformation is key for refinement.
    new_coords = np.array([np.dot(inv_trans_mat, (coords[i] - origin_shift))
                           for i in range(len(result))])
    if anchor_atom_index:
        offset = new_coords[anchor_atom_index] - anchor_atom_coords
        new_coords = new_coords - offset
    for i, coords in zip(result, new_coords):
        i.frac_coords = coords % 1

    if result != structure:
        logger.info(f"Structure is refined to the one with point group "
                    f"{symmetrizer.point_group}.")
        return result
    else:
        logger.info(f"Refined structure is same as before, so do nothing.")
